unit uHashItem;
{$INCLUDE 'vfy.inc'}

interface

uses
  SysUtils, Classes, Windows,
  uPCRE, uFiles, uSystem,
  uCompatibility, uProgress;

type
  THashItem = class;

  EHashItemError = class(EConvertError);

  THashValue = AnsiString;

  THashState = (hsOK, hsNew, hsChanged, hsDeleted, hsCopied, hsRenamed);
  THashStates = set of THashState;

  THashItem = class
    private
      fBasePath: string;
      fPath: string;
      fSize: int64;
      fDate: TDateTime;
      fHash: THashValue;
      fState: THashState;
      fRelated: THashItem;
    protected
      function GetAsString: string; virtual;
      procedure SetAsString(Value: string); virtual;
      function EncodeHash(const HashValue: THashValue): string; virtual;
      function DecodeHash(const StringValue: string): THashValue; virtual;
      function GetFileName: string; virtual;
    public
      class function GetFileDate(const FileName: string): TDateTime;
      class function GetFileSize(const FileName: string): int64;
      constructor Create; virtual;
      destructor Destroy; override;
      procedure Clear; virtual;
      procedure Assign(Source: THashItem); virtual;
      function ReadFromFile(const ABasePath, APath: string; OnProgress: TProgressEvent = nil {$IFDEF CASESENSITIVESUPPORT} ; CaseSensitive: boolean = False {$ENDIF} ): boolean; virtual;
      function BuildOutput(Detailed: boolean): string;
      function MatchOutput(Detailed: boolean): string;
      function DupesOutput(Detailed: boolean): string;
      property BasePath: string read fBasePath write fBasePath;
      property Path: string read fPath write fPath;
      property Size: int64 read fSize write fSize;
      property Date: TDateTime read fDate write fDate;
      property Hash: THashValue read fHash write fHash;
      property State: THashState read fState write fState;
      property Related: THashItem read fRelated write fRelated;
      property FileName: string read GetFileName;
      property AsString: string read GetAsString write SetAsString;
    end;

function HashItemCompareByPath(Item1, Item2: Pointer): integer;
function HashItemCompareByHash(Item1, Item2: Pointer): integer;

implementation

uses
  {$IFDEF DELPHI7_UP} RTLConsts, {$ELSE} {$IFNDEF FPC} Consts, {$ENDIF} {$ENDIF}
  DCPmd4, SynaCode;

type
  THash = TDCP_md4;

{$IFDEF FPC}
const SFOpenError = 'Cannot open file %s';
{$ENDIF}

const READ_BUFFER_SIZE = 8 * 1024 * 1024;

var ReadBuffer: Pointer;

{$IFDEF LARGEFILES}
const READ_BUFFER_ALIGN = 65536;

var ReadBufferAllocated: Pointer;
{$ENDIF}

var ItemAsStringRegExp: TRegExp = nil;

const
  BLANK_CHARS = [#0, #9, #10, #13, ' '];
  HASH_CHARS = ['0'..'9', 'a'..'z', 'A'..'Z', '+', '/', '='];
  HASH_CHARS_REGEXP = '[0-9a-zA-Z+/=]';

function HashItemCompareByPath(Item1, Item2: Pointer): integer;
var It1, It2: THashItem;
begin
  It1 := THashItem(Item1);
  It2 := THashItem(Item2);
  Result := AnsiCompareText(It1.Path, It2.Path);
end;

function HashItemCompareByHash(Item1, Item2: Pointer): integer;
var It1, It2: THashItem;
begin
  It1 := THashItem(Item1);
  It2 := THashItem(Item2);
  if It1.Hash = It2.Hash then
    if It1.Size = It2.Size then
      Result := 0
    else if It1.Size < It2.Size then
      Result := -1
    else
      Result := 1
  else if It1.Hash < It2.Hash then
    Result := -1
  else
    Result := 1;
end;

{ THashItem }

constructor THashItem.Create;
begin
  inherited Create;
  Clear;
end;

destructor THashItem.Destroy;
begin
  inherited;
end;

procedure THashItem.Clear;
begin
  BasePath := '';
  Path := '';
  Size := -1;
  Date := 0;
  Hash := '';
  State := hsOK;
  Related := nil;
end;

procedure THashItem.Assign(Source: THashItem);
begin
  BasePath := Source.BasePath;
  Path := Source.Path;
  Size := Source.Size;
  Date := Source.Date;
  Hash := Source.Hash;
  State := Source.State;
  Related := Source.Related;
end;

function THashItem.ReadFromFile(const ABasePath, APath: string; OnProgress: TProgressEvent {$IFDEF CASESENSITIVESUPPORT} ; CaseSensitive: boolean {$ENDIF} ): boolean;
{$UNDEF __HASHITEM_WINAPI}
{$IFDEF MSWINDOWS}
  {$IFDEF LARGEFILES} {$DEFINE __HASHITEM_WINAPI} {$ENDIF}
  {$IFDEF CASESENSITIVESUPPORT} {$DEFINE __HASHITEM_WINAPI} {$ENDIF}
{$ENDIF}
var TheHash: THash;
    TheFile: THandleStream;
    Count: integer;
    FileSize, TotalSize: int64;
    {$IFDEF __HASHITEM_WINAPI}
    CreateFileFlags: DWORD;
    Handle: THandle;
    {$ENDIF}
    Digest: AnsiString;
begin
  Clear;
  Self.BasePath := ABasePath;
  Self.Path := APath;
  TotalSize := GetFileSize(FileName);
  {$IFDEF __HASHITEM_WINAPI}
    CreateFileFlags := FILE_ATTRIBUTE_NORMAL;
    {$IFDEF LARGEFILES}
    if TotalSize >= 2145300000 then
      CreateFileFlags := CreateFileFlags or FILE_FLAG_NO_BUFFERING;
    {$ENDIF}
    {$IFDEF CASESENSITIVESUPPORT}
    if CaseSensitive then
      CreateFileFlags := CreateFileFlags or FILE_FLAG_POSIX_SEMANTICS;
    {$ENDIF}
    Handle := CreateFile(PChar(FileName), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, CreateFileFlags, 0);
    if Handle = INVALID_HANDLE_VALUE then
      {$IFDEF DELPHI2009_UP}
      Raise EFOpenError.CreateResFmt(@SFOpenErrorEx, [ExpandFileName(FileName), SysErrorMessage(GetLastError)]);
      {$ELSE}
        {$IFDEF FPC}
        Raise EFOpenError.CreateFmt(SFOpenError, [FileName]);
        {$ELSE}
        Raise EFOpenError.CreateResFmt(@SFOpenError, [FileName]);
        {$ENDIF}
      {$ENDIF}
    TheFile := THandleStream.Create(Handle);
  {$ELSE}
  TheFile := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  {$ENDIF}
  try
    FileSize := 0;
    TheHash := THash.Create(nil);
    try
      TheHash.Init;
      repeat
        Count := TheFile.Read(ReadBuffer^, READ_BUFFER_SIZE);
        if Count > 0 then
          begin
          FileSize := FileSize + Count;
          TheHash.Update(ReadBuffer^, Count);
          if Assigned(OnProgress) then
            OnProgress(Self, FileSize, TotalSize);
          end;
      until Count = 0;
      SetLength(Digest, TheHash.GetHashSize div 8);
      TheHash.Final(Digest[1]);
      Self.Hash := EncodeBase64(Digest);
      Self.Size := TotalSize;
      Self.Date := GetFileDate(FileName); //GetFileDateTime(FileName, DummyDate1, Self.fDate, DummyDate2); // Self.Date := FileGetDateTime(FileName);
    finally
      TheHash.Free;
      end;
  finally
    TheFile.Free;
    {$IFDEF __HASHITEM_WINAPI}
      CloseHandle(Handle);
    {$ENDIF}
    end;
  Result := Size >= 0;
{$UNDEF __HASHITEM_WINAPI}
end;

function THashItem.GetAsString: string;
var DateAsString, SizeAsString: string;
    i, iMin: integer;
begin
  if Self.Date > 0 then
    DateAsString := FormatDateTime('yyyy"-"mm"-"dd"T"hh":"nn":"ss.zzz', Self.Date)
  else
    DateAsString := '';
  SizeAsString := IntToStr(Size);
  if Size < 0 then
    iMin := 4
  else
    iMin := 3;
  i := Length(SizeAsString);
  while i > iMin do
    begin
    Dec(i, 3);
    Insert(',', SizeAsString, i+1);
    end;
  Result := Format('%-90s %s %16s %s', [AnsiQuotedStr(Path, '"'), DateAsString, SizeAsString, EncodeHash(Hash)]);
end;

procedure THashItem.SetAsString(Value: string);
const INVALID_INPUT = 'Invalid AsString input.';
const REGEXP_FILENAME_PART = '"(?P<PATH>[^"]+)"';
      REGEXP_YEAR_PART = '(?P<YEAR>[0-9]{4})';
      REGEXP_MONTH_PART = '(?P<MONTH>[0-9]{2})';
      REGEXP_DAY_PART = '(?P<DAY>[0-9]{2})';
      REGEXP_HOUR_PART = '(?P<HOUR>[0-9]{2})';
      REGEXP_MINUTE_PART = '(?P<MINUTE>[0-9]{2})';
      REGEXP_SECOND_PART = '(?P<SECOND>[0-9]{2})';
      REGEXP_MILLISECOND_PART = '(?P<MILLISECOND>[0-9]{3})';
      REGEXP_DATE_PART = '(?P<DATE>' + REGEXP_YEAR_PART + '-' + REGEXP_MONTH_PART + '-' + REGEXP_DAY_PART + 'T' + REGEXP_HOUR_PART + ':' + REGEXP_MINUTE_PART + ':' + REGEXP_SECOND_PART + '(?:\.' + REGEXP_MILLISECOND_PART + ')?\s+)?';
      REGEXP_SIZE_PART = '(?P<SIZE>-?[0-9]+(?:,[0-9]+)*)';
      REGEXP_HASH_PART = '(?P<HASH>' + HASH_CHARS_REGEXP + '+)';
      REGEXP_ASSTRING = '^\s*' + REGEXP_FILENAME_PART + '\s+' + REGEXP_DATE_PART + REGEXP_SIZE_PART + '\s+' + REGEXP_HASH_PART + '\s*$';
var Dummy: string;
begin
  Clear;
  {$IFDEF UNICODE}
  if Value <> '' then
    if Value[1] = #$feff then
      Value := Copy(Value, 2, MaxInt);
  {$ENDIF}
  if ItemAsStringRegExp = nil then
    ItemAsStringRegExp := RegExCreate(REGEXP_ASSTRING);
  if not ItemAsStringRegExp.Match(Value) then
    Raise EHashItemError.Create(INVALID_INPUT)
  else
    begin
    Self.Path := ItemAsStringRegExp.SubexpressionByName('PATH');
    if ItemAsStringRegExp.SubexpressionByName('DATE', Dummy) and (Dummy <> '') then
      Self.Date := EncodeDate(StrToInt(ItemAsStringRegExp.SubexpressionByName('YEAR')), StrToInt(ItemAsStringRegExp.SubexpressionByName('MONTH')), StrToInt(ItemAsStringRegExp.SubexpressionByName('DAY')))
                 + EncodeTime(StrToInt(ItemAsStringRegExp.SubexpressionByName('HOUR')), StrToInt(ItemAsStringRegExp.SubexpressionByName('MINUTE')), StrToInt(ItemAsStringRegExp.SubexpressionByName('SECOND')), StrToIntDef(ItemAsStringRegExp.SubexpressionByName('MILLISECOND'), 0));
    Self.Size := StrToInt64(StringReplace(ItemAsStringRegExp.SubexpressionByName('SIZE'), ',', '', [rfReplaceAll]));
    Self.Hash := DecodeHash(ItemAsStringRegExp.SubexpressionByName('HASH'));
    end;
end;

function THashItem.EncodeHash(const HashValue: THashValue): string;
begin
  Result := string(HashValue);
end;

function THashItem.DecodeHash(const StringValue: string): THashValue;
begin
  Result := THashValue(StringValue);
end;

class function THashItem.GetFileDate(const FileName: string): TDateTime;
var
  DummyDate1, DummyDate2: TDateTime;
begin
  uSystem.GetFileDateTime(FileName, DummyDate1, Result, DummyDate2);
end;

class function THashItem.GetFileSize(const FileName: string): int64;
begin
  Result := uSystem.GetFileSize(FileName);
end;

function THashItem.GetFileName: string;
begin
  Result := BasePath + Path;
end;

function THashItem.BuildOutput(Detailed: boolean): string;
begin
  if Detailed then
    Result := AsString
  else
    Result := AnsiQuotedStr(Path, '"');
end;

function THashItem.MatchOutput(Detailed: boolean): string;
const
  RES_MATCHED = 'MATCHED'#9'%s';
  RES_NEW = 'ADDED'#9'%s';
  RES_CHANGED = 'CHANGED'#9'%s';
  RES_DELETED = 'DELETED'#9'%s';
  RES_COPIED = 'COPIED'#9'%s'#9'from'#9'%s';
  RES_COPIED_X = 'COPIED'#9'%s';
  RES_RENAMED = 'RENAMED'#9'%s'#9'from'#9'%s';
  RES_RENAMED_X = 'RENAMED'#9'%s';
var
  Msg: string;
begin
  Result := '';
  if Detailed then
    Msg := AsString
  else
    Msg := AnsiQuotedStr(Path, '"');
  case State of
    hsOK:
      Result := Format(RES_MATCHED, [Msg]);
    hsNew:
      Result := Format(RES_NEW, [Msg]);
    hsChanged:
      Result := Format(RES_CHANGED, [Msg]);
    hsDeleted:
      Result := Format(RES_DELETED, [Msg]);
    hsCopied:
      if Related = nil then
        Result := Format(RES_COPIED_X, [Msg]) // This should never happen
      else
        Result := Format(RES_COPIED, [Msg, AnsiQuotedStr(Related.Path, '"')]);
    hsRenamed:
      if Related = nil then
        Result := Format(RES_RENAMED_X, [Msg]) // This should never happen
      else
        Result := Format(RES_RENAMED, [Msg, AnsiQuotedStr(Related.Path, '"')]);
    end;
end;

function THashItem.DupesOutput(Detailed: boolean): string;
begin
  Result := MatchOutput(Detailed);
end;

initialization
  {$IFDEF LARGEFILES}
  GetMem(ReadBufferAllocated, READ_BUFFER_SIZE + READ_BUFFER_ALIGN);
  ReadBuffer := Pointer(NativeUInt(ReadBufferAllocated) + READ_BUFFER_ALIGN - (NativeUInt(ReadBufferAllocated) mod READ_BUFFER_ALIGN));
  {$ELSE}
  GetMem(ReadBuffer, READ_BUFFER_SIZE);
  {$ENDIF}
  ItemAsStringRegExp := nil;

finalization
  {$IFDEF LARGEFILES}
  FreeMem(ReadBufferAllocated);
  ReadBufferAllocated := nil;
  {$ELSE}
  FreeMem(ReadBuffer);
  {$ENDIF}
  ReadBuffer := nil;
  RegExFreeAndNil(ItemAsStringRegExp);

end.
