unit uVerify;
{$INCLUDE 'vfy.inc'}

interface

uses
  SysUtils, Classes, Windows,
  uCompatibility, uSystem, uFiles, uPCRE,
  uHashList, uHashItem, uVerifyState;

type
  TVerify = class;

  TProcessFileFunction = function (const BasePath, Path: string; Tag: integer): integer of object;
  TAfterProcessFunction = procedure (var ProcessResult: integer; Tag: integer) of object;

  TIncludeExcludeMask = (iemInclude, iemExclude, iemIncludeRegExp, iemExcludeRegExp);
  TVerifyCommand = (vcDefault, vcVerify, vcBuild, vcUpdate, vcMatch, vcDupes, vcDupes1);

  TStateChangeEvent = procedure(Sender: TVerify; State: TVerifyState; const Info: TStateChangeObject) of object;
  TProgressEvent = procedure(Sender: TVerify; const Position, Max: int64) of object;

  TVerify = class(TObject)
    private
      fVerifyCommand: TVerifyCommand;
      fHashesFileName: string;
      fOutputFileName: string;
      fCacheFileName: string;
      fMatchListFileName: string;
      fBaseDirectories: TStringList;
      fRecursive: boolean;
      fJunctions: boolean;
      fHashList: THashList;
      fAuxHashList: THashList;
      fCache: THashList;
      fShowProgress: boolean;
      fIncludeExcludeList: TStringList;
      {$IFDEF CASESENSITIVESUPPORT}
      fCaseSensitive: boolean;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
      fWow64RedirectionOldValueDONOTCHANGE: Pointer;
      fDisableWow64: boolean;
      {$ENDIF}
      fUseTimestamp: boolean;
      fTimestampPrecision: integer;
      fHideStates: THashStates;
      fOnStateChange: TStateChangeEvent;
      fOnProgress: TProgressEvent;
      fAborted: boolean;
      fHashFileStream: TTextStream;
      fBaseDirectoriesRoot: string;
    fSaveFullPaths: boolean;
    protected
      procedure DoStateChange(State: TVerifyState; Info: TStateChangeObject = nil; FreeInfo: boolean = True); virtual;
      procedure DoProgress(const Position, Max: int64); virtual;
      function FixUpDirectory(const Dir: string): string; virtual;
      procedure FixUpBaseDirectories; virtual;
      procedure SeparateBaseDir(const Dir: string; out BaseDir, Path: string);
      procedure HashingProgress(Sender: TObject; Done, Total: int64); virtual;
      function ErrorInFile(E: Exception; const BasePath, Path: string): integer; virtual;
      procedure DeleteOutputFile;
      function ProcessFiles(const BasePath, Path: string; ProcessFileFunction: TProcessFileFunction; Tag: integer): integer; virtual;
      function ProcessFile_Build(const BasePath, Path: string; Tag: integer): integer; virtual;
      function ProcessFile_Verify(const BasePath, Path: string; Tag: integer): integer; virtual;
      function ProcessFile_Update(const BasePath, Path: string; Tag: integer): integer; virtual;
      procedure AfterProcess_Verify(var ProcessResult: integer; Tag: integer); virtual;
      procedure AfterProcess_Update(var ProcessResult: integer; Tag: integer); virtual;
      function InternalVerify(ProcessFileFunction: TProcessFileFunction; AfterProcessFunction: TAfterProcessFunction; Tag: integer): integer; virtual;
      function InternalMatch: integer; virtual;
      procedure ReadHashItem(HashItem: THashItem; const BasePath, Path: string);
      procedure ApplyIncludeExcludeToHashList(HashList: THashList); virtual;
      function IsFileWanted(const Path: string): boolean; virtual;
      procedure CheckControlEvent; virtual;
      procedure LoadHashFile(const FileName: string; HashList: THashList); virtual;
      {$IFDEF CASESENSITIVESUPPORT}
      function FindFirst(const Path: string; Attr: Integer; var SR: TSearchRec): Integer; virtual;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
        procedure ChangeWow64Redirection(Active: boolean); virtual;
        {$IFNDEF OVERRIDEWOW64GLOBALLY}
          {$IFDEF DELPHI6_UP}
            {$MESSAGE ERROR 'OVERRIDEWOW64 is not done yet.'}
          {$ELSE}
            'OVERRIDEWOW64 is not done yet.'
          {$ENDIF}
        {$ENDIF}
      {$ENDIF}
      property HashFileStream: TTextStream read fHashFileStream;
      property AuxHashList: THashList read fAuxHashList;
      property Cache: THashList read fCache;
      property IncludeExcludeList: TStringList read fIncludeExcludeList;
      property Aborted: boolean read fAborted write fAborted;
    public
      constructor Create; virtual;
      destructor Destroy; override;
      procedure Clear; virtual;
      procedure Abort; virtual;
      function Build: integer; virtual;
      function Verify: integer; virtual;
      function Update: integer; virtual;
      function Match: integer; virtual;
      function Dupes(OneFileOnly: boolean): integer; virtual;
      procedure ClearIncludeExcludeMasks; virtual;
      procedure AddIncludeExcludeMask(const Mask: string; MaskType: TIncludeExcludeMask); virtual;
      property VerifyCommand: TVerifyCommand read fVerifyCommand write fVerifyCommand;
      property HashesFileName: string read fHashesFileName write fHashesFileName;
      property OutputFileName: string read fOutputFileName write fOutputFileName;
      property CacheFileName: string read fCacheFileName write fCacheFileName;
      property MatchListFileName: string read fMatchListFileName write fMatchListFileName;
      property BaseDirectories: TStringList read fBaseDirectories;
      property BaseDirectoriesRoot: string read fBaseDirectoriesRoot write fBaseDirectoriesRoot;
      property Recursive: boolean read fRecursive write fRecursive;
      property ShowProgress: boolean read fShowProgress write fShowProgress;
      property Junctions: boolean read fJunctions write fJunctions;
      {$IFDEF CASESENSITIVESUPPORT}
      property CaseSensitive: boolean read fCaseSensitive write fCaseSensitive;
      {$ENDIF}
      {$IFDEF OVERRIDEWOW64}
      property DisableWow64: boolean read fDisableWow64 write fDisableWow64;
      {$ENDIF}
      property SaveFullPaths: boolean read fSaveFullPaths write fSaveFullPaths;
      property UseTimestamp: boolean read fUseTimestamp write fUseTimestamp;
      property TimestampPrecision: integer read fTimestampPrecision write fTimestampPrecision; // Milliseconds
      property HideStates: THashStates read fHideStates write fHideStates;
      property OnStateChange: TStateChangeEvent read fOnStateChange write fOnStateChange;
      property OnProgress: TProgressEvent read fOnProgress write fOnProgress;
      property HashList: THashList read fHashList;
    end;

const
  EXITCODE_OK = 0;
  EXITCODE_CHANGED_FILES = 1;
  EXITCODE_UNKNOWN_COMMAND = 101;
  EXITCODE_NO_DATA = 102;
  EXITCODE_ERROR_IN_FILES = 103;
  EXITCODE_BREAK = 253;
  EXITCODE_SYNTAX = 254;

implementation

{$IFDEF OVERRIDEWOW64GLOBALLY}
  {$IFDEF DELPHI6_UP}
    {$MESSAGE HINT 'Using OVERRIDEWOW64GLOBALLY. Don''t use LoadLibrary during Build/Verify/Update!'}
  {$ENDIF}
{$ENDIF}

const
  INCEXCTYPE_INCLUDE = 'I';
  INCEXCTYPE_EXCLUDE = 'X';

const
  TAG_BUILD = 1;
  TAG_VERIFY = 2;
  TAG_VERIFYFAST = 3;
  TAG_UPDATE = 4;
  TAG_UPDATEFAST = 5;

type
  EDelayedAbort = class(EAbort);

{ TVerify }

constructor TVerify.Create;
begin
  fBaseDirectories := TStringList.Create;
  fHashList := THashList.Create;
  fAuxHashList := THashList.Create;
  fCache := THashList.Create;
  fIncludeExcludeList := TStringList.Create;
  {$IFDEF OVERRIDEWOW64}
  fWow64RedirectionOldValueDONOTCHANGE := nil;
  {$ENDIF}
  inherited;
end;

procedure TVerify.DeleteOutputFile;
begin
  if OutputFileName <> '' then
    if FileExists(OutputFileName) then
      SysUtils.DeleteFile(OutputFileName);
end;

destructor TVerify.Destroy;
begin
  Clear;
  FreeAndNil(fBaseDirectories);
  FreeAndNil(fHashList);
  FreeAndNil(fCache);
  FreeAndNil(fAuxHashList);
  FreeAndNil(fIncludeExcludeList);
  inherited;
end;

procedure TVerify.Clear;
begin
  HashList.Clear;
  ClearIncludeExcludeMasks;
  VerifyCommand := vcDefault;
  HashesFileName := '';
  OutputFileName := '';
  MatchListFileName := '';
  BaseDirectories.Clear;
  BaseDirectoriesRoot := '';
  Recursive := True;
  Junctions := True;
  ShowProgress := True;
  {$IFDEF CASESENSITIVESUPPORT}
  CaseSensitive := False;
  {$ENDIF}
  {$IFDEF OVERRIDEWOW64}
  DisableWow64 := False;
  {$ENDIF}
  UseTimestamp := True;
  TimestampPrecision := 0;
  HideStates := [hsOK];
  Aborted := False;
end;

procedure TVerify.Abort;
begin
  Aborted := True;
end;

procedure TVerify.DoStateChange(State: TVerifyState; Info: TStateChangeObject; FreeInfo: boolean);
begin
  try
    if Assigned(OnStateChange) then
      OnStateChange(Self, State, Info);
  finally
    if Info <> nil then
      if FreeInfo then
        Info.Free;
    end;
end;

procedure TVerify.DoProgress(const Position, Max: int64);
begin
  if Assigned(OnProgress) then
    OnProgress(Self, Position, Max);
end;

procedure TVerify.FixUpBaseDirectories;
var i: integer;
begin
  for i := 0 to Pred(BaseDirectories.Count) do
    BaseDirectories[i] := FixUpDirectory(BaseDirectories[i]);
end;

function TVerify.FixUpDirectory(const Dir: string): string;
begin
  if Dir = '' then
    Result := IncludeTrailingPathDelimiter(GetCurrentDir)
  else
    Result := IncludeTrailingPathDelimiter(ExpandFileName(Dir));
  {$IFDEF LONGPATHS}
  if Result <> '' then
    if (Copy(Result, 1, 2) = '\\') and (Copy(Result, 3, 2) <> '?\') then
      Result := '\\?\UNC\' + Copy(Result, 3, MaxInt)
    else
      Result := '\\?\' + Result;
  {$ENDIF}
end;

function TVerify.ErrorInFile(E: Exception; const BasePath, Path: string): integer;
begin
  if E is EAbort then
    Result := EXITCODE_BREAK
  else
    begin
    Result := EXITCODE_ERROR_IN_FILES;
    DoStateChange(vstFileError, TStateChangeErrorInFile.Create(E, BasePath + Path));
    end;
end;

procedure TVerify.LoadHashFile(const FileName: string; HashList: THashList);
begin
  HashList.Clear;
  if ShowProgress then
    DoStateChange(vstLoadingHashFile, TStateChangeFilename.Create(FileName));
  HashList.LoadFromFile(FileName, '', HashingProgress);
  ApplyIncludeExcludeToHashList(HashList);
  if ShowProgress then
    DoStateChange(vstProgressDone);
end;

function TVerify.Build: integer;
var
  i, Code: integer;
  BaseDir, Path: string;
begin
  Result := EXITCODE_NO_DATA;
  HashList.Clear;
  AuxHashList.Clear;
  Cache.Clear;
  if CacheFileName <> '' then
    begin
    LoadHashFile(CacheFileName, Cache);
    Cache.SortByPath;
    end;
  DeleteOutputFile;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  if DisableWow64 then
    ChangeWow64Redirection(False);
  {$ENDIF}
  try
    fHashFileStream := TTextStream.Create(HashesFileName, fmCreate);
    try
      FreeAndNil(fHashFileStream);
      fHashFileStream := TTextStream.Create(HashesFileName, fmOpenReadWrite or fmShareDenyWrite);
      if FileExists(HashesFileName) then
        SysUtils.DeleteFile(HashesFileName);
      FixUpBaseDirectories;
      try
        for i := 0 to Pred(BaseDirectories.Count) do
          begin
          SeparateBaseDir(BaseDirectories[i], BaseDir, Path);
          Code := ProcessFiles(BaseDir, Path, ProcessFile_Build, TAG_BUILD);
          if Result = EXITCODE_NO_DATA then
            Result := Code
          else if Code <> EXITCODE_OK then
            Result := Code;
          end;
      except
        on E: EDelayedAbort do
          Result := EXITCODE_BREAK
        end;
      FreeAndNil(fHashFileStream);
      if Result <> EXITCODE_BREAK then
        begin
        if ShowProgress then
          DoStateChange(vstSortingResults);
        HashList.SortByPath;
        HashList.SaveToFile(HashesFileName);
        end;
    finally
      FreeAndNil(fHashFileStream);
      end;
  finally
    {$IFDEF OVERRIDEWOW64GLOBALLY}
    if DisableWow64 then
      ChangeWow64Redirection(True);
    {$ENDIF}
    end;
end;

function TVerify.ProcessFile_Build(const BasePath, Path: string; Tag: integer): integer;
var
  HashItem: THashItem;
begin
  Result := EXITCODE_OK;
  try
    HashItem := THashItem.Create;
    try
      ReadHashItem(HashItem, BasePath, Path);
      HashItem.State := hsNew;
      HashList.Add(HashItem);
      {$IFDEF SHOWRESULTSIMMEDIATELY}
      HashFileStream.WriteLine(HashItem.BuildOutput(True));
      {$ENDIF}
      if ShowProgress then
        DoStateChange(vstBuildOutput, TStateChangeHashItem.Create(HashItem));
    except
      HashItem.Free;
      Raise;
      end;
  except
    on E: EAbort do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
end;

function TVerify.InternalVerify(ProcessFileFunction: TProcessFileFunction; AfterProcessFunction: TAfterProcessFunction; Tag: integer): integer;
var
  i, Code: integer;
  BaseDir, Path: string;
  HashItem: THashItem;
begin
  Result := EXITCODE_NO_DATA;
  HashList.Clear;
  AuxHashList.Clear;
  Cache.Clear;
  if CacheFileName <> '' then
    begin
    LoadHashFile(CacheFileName, Cache);
    Cache.SortByPath;
    end;
  DeleteOutputFile;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  if DisableWow64 then
    ChangeWow64Redirection(False);
  try
  {$ENDIF}
  FixUpBaseDirectories;
  LoadHashFile(HashesFileName, HashList);
  if ShowProgress then
    DoStateChange(vstSortingHashlist);
  HashList.SortByPath;
  for i := 0 to Pred(HashList.Count) do
    begin
    HashItem := HashList[i];
    HashItem.State := hsOK;
    AuxHashList.AddCopy(HashItem);
    HashItem.State := hsDeleted;
    end;
  try
    for i := 0 to Pred(BaseDirectories.Count) do
      begin
      SeparateBaseDir(BaseDirectories[i], BaseDir, Path);
      Code := ProcessFiles(BaseDir, Path, ProcessFileFunction, Tag);
      if Result = EXITCODE_NO_DATA then
        Result := Code
      else if Code <> EXITCODE_OK then
        Result := Code;
      end;
  except
    on E: EDelayedAbort do
      Result := EXITCODE_BREAK;
    end;
  if Assigned(AfterProcessFunction) then
    AfterProcessFunction(Result, Tag);
  if (Result = EXITCODE_OK) or (Result = EXITCODE_CHANGED_FILES) then
    Result := InternalMatch;
  {$IFDEF OVERRIDEWOW64GLOBALLY}
  finally
    if DisableWow64 then
      ChangeWow64Redirection(True);
    end;
  {$ENDIF}
end;

function TVerify.Verify: integer;
begin
  Result := InternalVerify(ProcessFile_Verify, AfterProcess_Verify, TAG_VERIFY);
end;

function TVerify.ProcessFile_Verify(const BasePath, Path: string; Tag: integer): integer;
const
  MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000;
var
  HashItem: THashItem;
  HashListFileName: string;
  Index: integer;
begin
  Result := EXITCODE_OK;
  try
    if SaveFullPaths then
      HashListFileName := GetRealFileName(BasePath + Path)
    else
      HashListFileName := Path;
    if UseTimestamp
       and HashList.FindByPath(HashListFileName, Index)
       and (Abs(HashList[Index].Date - THashItem.GetFileDate(BasePath + Path)) <= (TimestampPrecision * 1/MILLISECONDS_PER_DAY))
       and (HashList[Index].Size = THashItem.GetFileSize(BasePath + Path))
    then
      begin
      HashList[Index].State := hsOK;
      Result := EXITCODE_OK;
      end
    else
      begin
      HashItem := THashItem.Create;
      try
        ReadHashItem(HashItem, BasePath, Path);
        if HashList.FindByPath(Path, Index) then
          if (HashList[Index].Size = HashItem.Size) and (HashList[Index].Hash = HashItem.Hash) then
            begin
            HashList[Index].Assign(HashItem);
            HashList[Index].State := hsOK;
            FreeAndNil(HashItem);
            end
          else
            begin
            HashItem.State := hsChanged;
            //HashList[Index].Free; // Free will be done by THashList itself due to OwnObjects=True
            HashList[Index] := HashItem;
            HashItem := nil;
            Result := EXITCODE_CHANGED_FILES;
            end
        else
          begin
          HashItem.State := hsNew;
          HashList.Insert(Index, HashItem);
          HashItem := nil;
          Result := EXITCODE_CHANGED_FILES;
          end;
      except
        HashItem.Free;
        Raise;
        end;
      end;
  except
    on E: EAbort do
      Raise;
    on E: Exception do
      Result := ErrorInFile(E, BasePath, Path);
    end;
end;

procedure TVerify.ReadHashItem(HashItem: THashItem; const BasePath, Path: string);
var
  CachedFileName: string;
  Index: integer;
begin
  if SaveFullPaths or ((Cache <> nil) and (Cache.Count > 0)) then
    CachedFileName := GetRealFileName(BasePath + Path)
  else
    CachedFileName := '';
  if (Cache <> nil) and (Cache.Count > 0) and (CachedFileName <> '') and Cache.FindByPath(CachedFileName, Index) then
    begin
    HashItem.Assign(Cache[Index]);
    if SaveFullPaths and (CachedFileName <> '') then
      begin
      HashItem.BasePath := '';
      HashItem.Path := CachedFileName;
      end
    else
      begin
      HashItem.BasePath := BasePath;
      HashItem.Path := Path;
      end;
    end
  else
    begin
    HashItem.ReadFromFile(BasePath, Path, HashingProgress {$IFDEF CASESENSITIVESUPPORT} , CaseSensitive {$ENDIF} );
    if SaveFullPaths and (CachedFileName <> '') then
      begin
      HashItem.BasePath := '';
      HashItem.Path := CachedFileName;
      end;
    end;
end;

procedure TVerify.SeparateBaseDir(const Dir: string; out BaseDir, Path: string);
var
  ix, RootLen: integer;
  Root: string;
begin
  ix := Pos('|', Dir);
  if ix > 0 then
    begin
    BaseDir := IncludeTrailingPathDelimiter(Copy(Dir, 1, Pred(ix)));
    if ix < Length(Dir) then
      if Dir[Succ(ix)] = '\' then
        Inc(ix);
    Path := Copy(Dir, Succ(ix), MaxInt);
    end
  else
    begin
    BaseDir := Dir;
    Path := '';
    if BaseDirectoriesRoot <> '' then
      begin
      Root := FixUpDirectory(BaseDirectoriesRoot);
      RootLen := Length(Root);
      if AnsiCompareText(Root, Copy(Dir, 1, Length(Root))) = 0 then
        begin
        BaseDir := IncludeTrailingPathDelimiter(Root);
        if RootLen < Length(Dir) then
          if Dir[Succ(RootLen)] = '\' then
            Inc(RootLen);
        Path := Copy(Dir, Succ(RootLen), MaxInt);
        end;
      end;
    end;
end;

procedure TVerify.AfterProcess_Verify(var ProcessResult: integer; Tag: integer);
begin
  HashList.DeleteByState(hsDeleted);
end;

function TVerify.Update: integer;
begin
  Result := InternalVerify(ProcessFile_Update, AfterProcess_Update, TAG_UPDATE);
end;

function TVerify.ProcessFile_Update(const BasePath, Path: string; Tag: integer): integer;
begin
  Result := ProcessFile_Verify(BasePath, Path, Tag);
end;

procedure TVerify.AfterProcess_Update(var ProcessResult: integer; Tag: integer);
begin
  AfterProcess_Verify(ProcessResult, Tag);
  HashList.SortByPath;
  HashList.SaveToFile(HashesFileName);
end;

function TVerify.InternalMatch: integer;
  // HashList - new (current) hashlist
  // AuxHashList - old hashlist
var
  i, Index: integer;
  Done, Total: int64;
  State: THashState;
  StateCount: array[THashState] of integer;
  OutputStream: TTextStream;

  procedure Step;
    begin
      Inc(Done);
      HashingProgress(Self, Done, Total);
    end;

begin
  Result := EXITCODE_OK;
  if ShowProgress then
    DoStateChange(vstComparingHashlists);
  Total := 2 * AuxHashList.Count + 2 * HashList.Count;
  Done := -1;
  Step;
  for i := 0 to Pred(AuxHashList.Count) do
    begin
    AuxHashList[i].State := hsDeleted;
    Step;
    end;
  // 1. Match by path
  AuxHashList.SortByPath;
  for i := 0 to Pred(HashList.Count) do
    begin
    if AuxHashList.FindByPath(HashList[i].Path, Index) then
      begin
      HashList[i].State := hsChanged;
      HashList[i].Related := AuxHashList[Index];
      while (Index < AuxHashList.Count) and (HashItemCompareByPath(HashList[i], AuxHashList[Index]) = 0) do
        begin
        if HashItemCompareByHash(HashList[i], AuxHashList[Index]) = 0 then
          begin
          if HashList[i].State <> hsOK then
            begin
            HashList[i].State := hsOK;
            HashList[i].Related := AuxHashList[Index];
            end;
          AuxHashList[Index].State := hsOK;
          end
        else
          begin
          if AuxHashList[Index].State = hsDeleted then
            AuxHashList[Index].State := hsChanged;
          end;
        Inc(Index);
        end;
      end
    else
      begin
      HashList[i].Related := nil;
      HashList[i].State := hsNew;
      end;
    Step;
    end;
  // 2. Match not-matched by hash
  AuxHashList.SortByHash;
  for i := 0 to Pred(HashList.Count) do
    begin
    if not (HashList[i].State in [hsOK, hsChanged]) then
      if AuxHashList.FindByHash(HashList[i].Hash, HashList[i].Size, Index) then
        begin
        HashList[i].State := hsCopied;
        HashList[i].Related := AuxHashList[Index];
        while (Index < AuxHashList.Count) and (HashItemCompareByHash(HashList[i], AuxHashList[Index]) = 0) do
          begin
          if AuxHashList[Index].State <> hsOK then
            begin
            AuxHashList[Index].State := hsRenamed;
            if (AuxHashList[Index].Related = nil) or (AuxHashList[Index] <> AuxHashList[Index].Related.Related) then
              begin
              HashList[i].Related := AuxHashList[Index];
              AuxHashList[Index].Related := HashList[i];
              end;
            end;
          Inc(Index);
          end;
        end;
      Step;
      end;
  // 3. Process missing or renamed files from the old hashes
  for i := 0 to Pred(AuxHashList.Count) do
    begin
    if AuxHashList[i].State = hsRenamed then
      if AuxHashList[i].Related <> nil then
        if AuxHashList[i].Related.State = hsCopied then
          AuxHashList[i].Related.State := hsRenamed
        else
          AuxHashList[i].State := hsDeleted;
    if AuxHashList[i].State = hsDeleted then
      HashList.AddCopy(AuxHashList[i]);
    end;
  // 4. Print out the results
  if ShowProgress then
    DoStateChange(vstSortingResults);
  HashList.SortByPath;
  if ShowProgress then
    DoStateChange(vstWritingResults);
  for State := Low(THashState) to High(THashState) do
    StateCount[State] := 0;
  OutputStream := nil;
  try
    for i := 0 to Pred(HashList.Count) do
      begin
      Inc(StateCount[HashList[i].State]);
      if not (HashList[i].State in HideStates) then
        begin
        if (Result = EXITCODE_OK) or (Result = EXITCODE_NO_DATA) then
          Result := EXITCODE_CHANGED_FILES;
        if OutputStream = nil then
          if OutputFileName <> '' then
            OutputStream := TTextStream.Create(OutputFileName, fmCreate);
        if OutputStream <> nil then
          OutputStream.WriteLine(HashList[i].MatchOutput(True));
        end;
      if ShowProgress then
        DoStateChange(vstMatchOutput, TStateChangeHashItem.Create(HashList[i]));
      end;
  finally
    FreeAndNil(OutputStream);
    end;
  if ShowProgress then
    DoStateChange(vstMatchStats, TStateChangeStatistics.Create(StateCount[hsOK], StateCount[hsNew], StateCount[hsChanged], StateCount[hsDeleted], StateCount[hsCopied], StateCount[hsRenamed]));
end;

function TVerify.Match: integer;
begin
  HashList.Clear;
  AuxHashList.Clear;
  DeleteOutputFile;
  LoadHashFile(HashesFileName, HashList);
  LoadHashFile(MatchListFileName, AuxHashList);
  Result := InternalMatch;
end;

function TVerify.Dupes(OneFileOnly: boolean): integer;
var
  SearchList: THashList;
  //Item: THashItem;
  i, Index: integer;
  Done, Total: int64;
  State: THashState;
  StateCount: array[THashState] of integer;
  OutputStream: TTextStream;

  procedure Step;
    begin
      Inc(Done);
      HashingProgress(Self, Done, Total);
    end;

begin
  Result := EXITCODE_OK;
  HashList.Clear;
  AuxHashList.Clear;
  DeleteOutputFile;
  LoadHashFile(HashesFileName, HashList);
  if OneFileOnly then
    SearchList := HashList
  else
    begin
    SearchList := AuxHashList;
    LoadHashFile(MatchListFileName, SearchList);
    end;
  SearchList.SortByHash;
  if ShowProgress then
    DoStateChange(vstLookingForDupes);
  Total := HashList.Count;
  Done := -1;
  Step;
  for State := Low(THashState) to High(THashState) do
    StateCount[State] := 0;
  if OutputFileName = '' then
    OutputStream := nil
  else
    OutputStream := TTextStream.Create(OutputFileName, fmCreate);
  try
    for i := 0 to Pred(HashList.Count) do
      begin
      HashList[i].State := hsNew;
      if SearchList.FindByHash(HashList[i].Hash, HashList[i].Size, Index) then
        repeat
          if HashList[i] <> SearchList[Index] then
            begin
            HashList[i].State := hsCopied;
            HashList[i].Related := SearchList[Index];
            if OutputStream <> nil then
              OutputStream.WriteLine(HashList[i].DupesOutput(True));
            if ShowProgress then
              DoStateChange(vstDupesOutput, TStateChangeHashItem.Create(HashList[i]));
            end;
          Inc(Index);
        until (Index >= SearchList.Count) or (HashItemCompareByHash(HashList[i], SearchList[Index]) <> 0);
      Inc(StateCount[HashList[i].State]);
      Step;
      end;
  finally
    FreeAndNil(OutputStream);
    end;
  if ShowProgress then
    DoStateChange(vstDupesStats, TStateChangeStatistics.Create(StateCount[hsOK], StateCount[hsNew], StateCount[hsChanged], StateCount[hsDeleted], StateCount[hsCopied], StateCount[hsRenamed]));
end;

{$IFDEF OVERRIDEWOW64}
procedure TVerify.ChangeWow64Redirection(Active: boolean);
var OK: boolean;
begin
  if not Is64BitWindows then
    Exit;
  if Active then
    if fWow64RedirectionOldValueDONOTCHANGE <> nil then
      OK := Wow64RevertWow64FsRedirection(fWow64RedirectionOldValueDONOTCHANGE)
    else
      OK := False
  else
    OK := Wow64DisableWow64FsRedirection(fWow64RedirectionOldValueDONOTCHANGE);
  if not OK then
    Raise EOSError.CreateFmt('Failed to change WoW64 redirection state to %d.', [Integer(Active)]);
end;
{$ENDIF}

{$IFDEF CASESENSITIVESUPPORT}
{$IFDEF DELPHI6_UP}
  {$WARN UNIT_PLATFORM OFF}
  {$WARN SYMBOL_PLATFORM OFF}
{$ENDIF}

{$IFNDEF DELPHI6_UP}
  // Delphi 5 has an incorrect declaration of Windows.FindFirstFileEx
  function FindFirstFileEx(lpFileName: PChar; fInfoLevelId: TFindexInfoLevels;
    lpFindFileData: Pointer; fSearchOp: TFindexSearchOps; lpSearchFilter: Pointer;
    dwAdditionalFlags: DWORD): THandle; stdcall; external kernel32 name 'FindFirstFileExA';
{$ENDIF}

  // Copied from SysUtils
  function FindMatchingFile(var F: TSearchRec): Integer;
  var
    LocalFileTime: TFileTime;
  begin
    with F do
    begin
      while FindData.dwFileAttributes and ExcludeAttr <> 0 do
        if not FindNextFile(FindHandle, FindData) then
        begin
          Result := GetLastError;
          Exit;
        end;
      FileTimeToLocalFileTime(FindData.ftLastWriteTime, LocalFileTime);
      FileTimeToDosDateTime(LocalFileTime, LongRec(Time).Hi,
        LongRec(Time).Lo);
      Size := FindData.nFileSizeLow;
      Attr := FindData.dwFileAttributes;
      Name := FindData.cFileName;
    end;
    Result := 0;
  end;

function TVerify.FindFirst(const Path: string; Attr: Integer; var SR: TSearchRec): Integer;
begin
  // This will only set-up the contents of SR
  Result := SysUtils.FindFirst(Path, Attr, SR);
  if CaseSensitive then
    begin
    SysUtils.FindClose(SR);
    // Now I will create the search handle with required parameters
    SR.FindHandle := FindFirstFileEx(PChar(Path), FindExInfoStandard, @SR.FindData, FindExSearchNameMatch, nil, FIND_FIRST_EX_CASE_SENSITIVE);
    if SR.FindHandle <> INVALID_HANDLE_VALUE then
      begin
      Result := FindMatchingFile(SR);
      if Result <> 0 then SysUtils.FindClose(SR);
      end
    else
      Result := GetLastError;
    end;
end;

{$IFDEF DELPHI6_UP}
  {$WARN UNIT_PLATFORM ON}
  {$WARN SYMBOL_PLATFORM ON}
{$ENDIF}
{$ENDIF}

function TVerify.ProcessFiles(const BasePath, Path: string; ProcessFileFunction: TProcessFileFunction; Tag: integer): integer;
const ITEM_IS_FILE = 0;
      ITEM_IS_DIRECTORY = 1;
var SR: TSearchRec;
    List: TStringList;
    i, Code: integer;
begin
  List := TStringList.Create;
  try
    if FindFirst(BasePath + Path + '*.*', faAnyFile, SR) = 0 then
      try
        repeat
          CheckControlEvent;
          {$IFDEF DELPHI7_UP}
            {$WARN SYMBOL_PLATFORM OFF}
          {$ENDIF}
          if (not Junctions) and Longbool(SR.FindData.dwFileAttributes and FILE_ATTRIBUTE_REPARSE_POINT) then
            begin
            // Skip
            end
          else
          if Longbool(SR.Attr and faDirectory) then
            begin
            if Recursive and (SR.Name <> '.') and (SR.Name <> '..') then
              List.AddObject(SR.Name, TObject(ITEM_IS_DIRECTORY));
            end
          {$IFNDEF DELPHI2007_UP}
          else if Longbool(SR.Attr and faVolumeID) then
            begin
            // Nothing to do
            end
          {$ENDIF}
          {$IFDEF DELPHI7_UP}
            {$WARN SYMBOL_PLATFORM ON}
          {$ENDIF}
          else
            begin
            List.AddObject(SR.Name, TObject(ITEM_IS_FILE));
            end;
        until FindNext(SR) <> 0;
      finally
        SysUtils.FindClose(SR);
        end;
    List.Sort;
    Result := EXITCODE_OK;
    for i := 0 to Pred(List.Count) do
      begin
      CheckControlEvent;
      Code := EXITCODE_OK;
      if List.Objects[i] = TObject(ITEM_IS_DIRECTORY) then
        Code := ProcessFiles(BasePath, Path + List[i] + '\', ProcessFileFunction, Tag)
      else if IsFileWanted(Path + List[i]) then
        begin
        if ShowProgress then
          DoStateChange(vstProcessingFile, TStateChangeFileName.Create(Path + List[i]));
        Code := ProcessFileFunction(BasePath, Path + List[i], Tag);
        if ShowProgress then
          DoStateChange(vstProgressDone);
        end;
      if Code <> EXITCODE_OK then
        Result := Code;
      end;
  finally
    List.Free;
    end;
end;

procedure TVerify.HashingProgress(Sender: TObject; Done, Total: int64);
begin
  CheckControlEvent;
  if ShowProgress then
    DoProgress(Done, Total);
end;

function CreateRegEx(const Pattern: string): TRegExp;
begin
  Result := RegExCreate(Pattern, [rcoSingleLine, rcoIgnoreCase]);
end;

function CreateRegExFromFileMask(Mask: string): TRegExp;
var i: integer;
    Pattern: string;
begin
  Mask := ExtractFileName(Mask); // Ignore path
  Pattern := '';
  for i := 1 to Length(Mask) do
    case Mask[i] of
      '*':
        Pattern := Pattern + '.*';
      '?':
        Pattern := Pattern + '.';
      ']':
        Pattern := Pattern + '\]';
      '^':
        Pattern := Pattern + '\^';
      '\':
        Pattern := Pattern + '\\';
      else
        Pattern := Pattern + '[' + Mask[i] + ']';
      end;
  Result := CreateRegEx('^' + Pattern + '$');
end;

procedure TVerify.ClearIncludeExcludeMasks;
var i: integer;
begin
  for i := 0 to Pred(IncludeExcludeList.Count) do
    RegExFree(TRegExp(IncludeExcludeList.Objects[i]));
  IncludeExcludeList.Clear;
end;

procedure TVerify.AddIncludeExcludeMask(const Mask: string; MaskType: TIncludeExcludeMask);
begin
  if IncludeExcludeList.Count = 0 then
    case MaskType of
      iemInclude, iemIncludeRegExp:
        IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE, CreateRegEx('.*'));
      iemExclude, iemExcludeRegExp:
        IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE, CreateRegEx('.*'));
      end;
  case MaskType of
    iemInclude:
      IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE + Mask, CreateRegExFromFileMask(Mask));
    iemExclude:
      IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE + Mask, CreateRegExFromFileMask(Mask));
    iemIncludeRegExp:
      IncludeExcludeList.AddObject(INCEXCTYPE_INCLUDE + Mask, CreateRegEx(Mask));
    iemExcludeRegExp:
      IncludeExcludeList.AddObject(INCEXCTYPE_EXCLUDE + Mask, CreateRegEx(Mask));
    end;
end;

procedure TVerify.ApplyIncludeExcludeToHashList(HashList: THashList);
var i: integer;
begin
  for i := Pred(HashList.Count) downto 0 do
    if not IsFileWanted(HashList[i].Path) then
      HashList.Delete(i);
end;

function TVerify.IsFileWanted(const Path: string): boolean;
var i: integer;
    s: string;
begin
  Result := True;
  for i := 0 to Pred(IncludeExcludeList.Count) do
    if TRegExp(IncludeExcludeList.Objects[i]).Match(Path) then
      begin
      s := IncludeExcludeList[i];
      if (s <> '') and (s[1] = INCEXCTYPE_INCLUDE) then
        Result := True
      else if (s <> '') and (s[1] = INCEXCTYPE_EXCLUDE) then
        Result := False;
      end;
end;

procedure TVerify.CheckControlEvent;
begin
  if Aborted then
    begin
    Aborted := False;
    Raise EDelayedAbort.Create('Task aborted.');
    end;
end;

end.
